#Copyright (C) 2020 Samira Maghool and Marco Cremonini
#
#This file is part of the multiagent epidemic simulator developed by the research project described in the paper titled "A multicomponent model of awareness for different categories of network epidemics", by Samira Maghool, Nahid Maleki-Jirsaraei, and Marco Cremonini.
#
#    The multiagent epidemic simulator is free software: you can  
#    redistribute it and/or modify it under the terms of the GNU General
#    Public License as published by the Free Software Foundation, either
#    version 3 of the License, or (at your option) any later version.
#
#    The multiagent epidemic simulator is distributed in the hope that it 
#    will be useful, but WITHOUT ANY WARRANTY; without even the implied 
#    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#    See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with the multiagent epidemic simulator.  
#    If not, see <https://www.gnu.org/licenses/>.
        
''' This module is for running the agent-based simulation'''
''' @author: samiramaghool@gmail.com>'''

''' Here, you'll find the codes for initializing, network configuration and Run the simulation'''


from os import sep
import NetworkSimulation
import SIRSAgent
import networkx as nx
from networkx import Graph, DiGraph
import random
import numpy as np
from plotting import PlotCreator

#paths
SOURCE_DIR = 'source'
RESULTS_DIR = 'results'
NET_IMAGES_DIR = 'netDraw'  
myName = "SIR"
TRIALS = 20
title = "Simulation of agent-based Coronavirus "


'''define simulation-specific constants:'''

SUSCEPTIBLE = 0
INCUBATING_INFECTED = 1
MILD_INFECTED = 2
ACUTE_INFECTED = 3
CONTAINED = 4
RECOVERED_MI = 5
RECOVERED_AI = 6

SIRSNet_NAME = 'SIRS_edgelist.txt'


'''simulation parameters'''

MAX_SIMULATION_TIME = 140.0
NODES = 1000.

PRINT_GIF = False


'''Retrieve Networks'''


SIRSNet = nx.powerlaw_cluster_graph(NODES, 3, 0.6)



def retrieveNet(path, netType):
    return read_edgelist(path, create_using=netType, nodetype=int)

def storeNet(net, path):
    write_edgelist(net, path)

    
def main():

    '''Node states initialization'''
    x = random.uniform(0, NODES)
    SIRSStates = setupInitialStates(SIRSNet, int(x), int(x+5), NetworkSimulation.INCUBATING_INFECTED, NetworkSimulation.SUSCEPTIBLE)
        
    '''parameters setting'''
   
        
    for i in range(0, SIRSNet.__len__()):      
           
        SIRSNet.nodes[i]['t_II'] = 0
        SIRSNet.nodes[i]['t_MI'] = 0
        SIRSNet.nodes[i]['t_AI'] = 0
        SIRSNet.nodes[i]['t_C'] = 0
        
        #q1 = 5
        #SIRSNet.nodes[i]['t_II_I'] = q1
        q1 = int(np.random.gamma(3., 1.))     
        t_min = 2
        t_max = 5

        if q1 == 0:    
           SIRSNet.nodes[i]['t_II_I'] = t_min      #random day for transition from II to I(AI or MI)
        elif q1 == 1:
           SIRSNet.nodes[i]['t_II_I'] = t_min+1
        elif q1 == 2:
            SIRSNet.nodes[i]['t_II_I'] = t_min+2
        elif q1 >= 3:
            SIRSNet.nodes[i]['t_II_I'] = t_min+5
            
        
        if q1<= 2:    
           SIRSNet.nodes[i]['t_II_I'] = 2      #random day for transition from II to I(AI or MI)
        elif q1 >= 5:
           SIRSNet.nodes[i]['t_II_I'] = 5
        else:
            SIRSNet.nodes[i]['t_II_I'] = q1




        
            
        q2 = random.randint (2, 7)
        SIRSNet.nodes[i]['t_MI_R'] = q2      #random day for transition from MI to R like normal flu


        
       
        q3 = int(np.random.gamma(3, 1.))         
        
        if q3 <= 2:
            SIRSNet.nodes[i]['t_AI_C'] = 2      #random day for transition from AI to C (T_d)
        elif q3 >= 7 :
            SIRSNet.nodes[i]['t_AI_C'] = 7
        else:
            SIRSNet.nodes[i]['t_AI_C'] = q3


      
            
        q4 = random.randint (2, 7)
        SIRSNet.nodes[i]['t_C_R/MI'] = q4      #random day for transition from C to R(MIDL CONTAINED)


        
        q5 = random.randint (14, 30)
        SIRSNet.nodes[i]['t_C_R/AI'] = q5      #random day for transition from C to R(ACUTE CONTAINED)
        

      
        q6 = random.randint (2, 7)   
        SIRSNet.nodes[i]['t_MI_C'] = q6      #random day for transition from MI to R like normal flu

    '''RUN SIMULATION'''
    simulation = NetworkSimulation.NetworkSimulation(SIRSNet, SIRSAgent, SIRSStates, 
                                                     RESULTS_DIR,
                                                     NET_IMAGES_DIR,
                                                     MAX_SIMULATION_TIME,
                                                     TRIALS)
    
    

    
    simulation.runSimulation()  #run rimulation
    
def setupInitialStates(Network, minRange, maxRange, specific_State, default_State):
   
    result1 = [default_State for i in range(0, minRange)]
    result2 = [specific_State for i in range(minRange, maxRange)]
    result3 = [default_State for i in range(maxRange, Network.__len__())]
    return result1 + result2 + result3


if __name__ == '__main__':
    main()
 
