#Copyright (C) 2020 Samira Maghool and Marco Cremonini
#
#This file is part of the multiagent epidemic simulator developed by the research project described in the paper titled "A multicomponent model of awareness for different categories of network epidemics", by Samira Maghool, Nahid Maleki-Jirsaraei, and Marco Cremonini.
#
#    The multiagent epidemic simulator is free software: you can  
#    redistribute it and/or modify it under the terms of the GNU General
#    Public License as published by the Free Software Foundation, either
#    version 3 of the License, or (at your option) any later version.
#
#    The multiagent epidemic simulator is distributed in the hope that it 
#    will be useful, but WITHOUT ANY WARRANTY; without even the implied 
#    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#    See the GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with the multiagent epidemic simulator.  
#    If not, see <https://www.gnu.org/licenses/>.
        
''' This module is for running the agent-based simulation'''
''' @author: samiramaghool@gmail.com>'''

'''
Module for the SIRSAgent class that can be subclassed by agents.

@Author:  <samiramaghool@gmail.com>
'''
from SimPy import Simulation as Sim
import random
import NetworkSimulation   #why this?in original form it it networkx
import math
SEED = 9154853443

'''shared parameters for SIRSAgents'''


P_S = 0.97              #infection probabality is 1-P_S (lambda)
P_II_AI = 0.2           # probability of moving from incubating infected (II) to acute infected (AI)
P_MI_C = 0              # probability of moving from Mild infected (MI) to Contained (C)
P_AI_R = 1.             # probability of moving from Acute infected (AI) to Recovered (R) after contained period

class SIRSAgent(Sim.Process):
    
    #class variables, shared between all instances of this class
    r = random.Random(SEED)
    
    TIMESTEP_DEFAULT = 1.0
    
    def __init__(self, SIRSState, initialiser, name='network_process'):   
        Sim.Process.__init__(self, name)
        self.SIRSState = SIRSState
        self.initialize(*initialiser)
        
        
    def initialize(self, id, sim, SIRSNet):       
        ''' this gets called automatically '''        
        self.id = id
        self.sim = sim
        self.SIRSNet = SIRSNet
        
    
    def getAllNodes(self):
        ''' restituisce tutti i nodi nella rete SIRS '''
        return self.SIRSNet.nodes()
    
    def getAllAgents(self, SIRSState=None):
        ''' restituisce tutti gli agenti di tipo SIRS '''
        neighs = self.getAllNodes()
        if SIRSState is not None:
            return [self.SIRSNet.node[n]['agent'] for n in neighs 
                      if self.SIRSNet.node[n]['agent'].SIRSState == SIRSState]
        else:
            return [self.SIRSNet.node[n]['agent'] for n in neighs] 

    def getNeighbouringAgents(self, SIRSState=None):
        ''' returns list of neighbours, but as agents, not nodes.
        so e.g. one can set result[0].state = INFECTED '''
        neighs = self.SIRSNet.neighbors(self.id)
        if SIRSState is not None:
            return [self.SIRSNet.node[n]['agent'] for n in neighs 
                      if self.SIRSNet.node[n]['agent'].SIRSState == SIRSState]
        else:
            return [self.SIRSNet.node[n]['agent'] for n in neighs]
        
    def getNeighbouringAgentsIter(self, SIRSState=None):
        '''same as getNeighbouringAgents, but returns generator expression, not list. '''
        neighs = self.SIRSNet.neighbors(self.id)
        if SIRSState is not None:
            return (self.SIRSNet.node[n]['agent'] for n in neighs 
                      if self.SIRSNet.node[n]['agent'].SIRSState == SIRSState)
        else:
            return (self.SIRSNet.node[n]['agent'] for n in neighs)
    
    def getNeighbouringNodes(self):
        ''' returns list of neighbours as nodes. 
        Call self.getAgent() on one of them to get the agent.'''
        return self.SIRSNet.neighbors(self.id)
   
    def getAgent(self, id):
        '''returns agent of specified ID.'''   
        return self.SIRSNet.node[id]['agent']
             
    def logTopoChange(self, action, node, node2=None):
        '''#TODO: test, add this to netlogger...'''
        print (action, node, node2)

   
             
        
    def Run(self):
        while True:
            
            if self.SIRSState == NetworkSimulation.SUSCEPTIBLE:
                self.mayBecomeIncubating_Infected()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT #wait a step
            elif self.SIRSState == NetworkSimulation.INCUBATING_INFECTED:
                self.mayBecomeInfected()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            elif self.SIRSState == NetworkSimulation.MILD_INFECTED:
                self.mayBecomeRecoveredORContained()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            elif self.SIRSState == NetworkSimulation.ACUTE_INFECTED:
                self.mayBecomeRecoveredORContained()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            elif self.SIRSState == NetworkSimulation.CONTAINED:
                self.mayBecomeRecoveredOrDead()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            elif self.SIRSState == NetworkSimulation.RECOVERED_MI:
                self.mayStayRecoveredOrDead()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            elif self.SIRSState == NetworkSimulation.RECOVERED_AI:
                self.mayStayRecoveredOrDead()
                yield Sim.hold, self, SIRSAgent.TIMESTEP_DEFAULT  #wait a step
            
    
            
    def mayBecomeIncubating_Infected(self):
          
            neighbours = self.getNeighbouringNodes()        
                                 
            for i in neighbours:           
                if self.SIRSNet.node[i]['agent'].SIRSState == NetworkSimulation.ACUTE_INFECTED:
                    if SIRSAgent.r.random() < 1- P_S:       
                        self.SIRSState = NetworkSimulation.INCUBATING_INFECTED
                        self.SIRSNet.nodes[self.id]['t_II'] = self.sim.now()
                elif self.SIRSNet.node[i]['agent'].SIRSState == NetworkSimulation.MILD_INFECTED:
                    if SIRSAgent.r.random() < 0.5*(1- P_S):       
                        self.SIRSState = NetworkSimulation.INCUBATING_INFECTED
                        self.SIRSNet.nodes[self.id]['t_II'] = self.sim.now()
                elif self.SIRSNet.node[i]['agent'].SIRSState == NetworkSimulation.INCUBATING_INFECTED:
                    if SIRSAgent.r.random() < 0.5*(1- P_S):       
                        self.SIRSState = NetworkSimulation.INCUBATING_INFECTED
                        self.SIRSNet.nodes[self.id]['t_II'] = self.sim.now()
                
                  


    def mayBecomeInfected(self):
       
        if self.sim.now() - self.SIRSNet.nodes[self.id]['t_II'] == self.SIRSNet.nodes[self.id]['t_II_I'] :
            
            if SIRSAgent.r.random() < P_II_AI :        
                self.SIRSState = NetworkSimulation.ACUTE_INFECTED
                self.SIRSNet.nodes[self.id]['t_AI'] = self.sim.now()
            else:
                self.SIRSState = NetworkSimulation.MILD_INFECTED
                self.SIRSNet.nodes[self.id]['t_MI'] = self.sim.now()
            
        

    def mayBecomeRecoveredORContained(self):
      
        if self.SIRSState == NetworkSimulation.MILD_INFECTED:
           if SIRSAgent.r.random() <= (1 - P_MI_C):                      
              if self.sim.now() - self.SIRSNet.nodes[self.id]['t_MI'] >= self.SIRSNet.nodes[self.id]['t_MI_R']:
                  self.SIRSState = NetworkSimulation.RECOVERED_MI
           else:         
              if self.sim.now() - self.SIRSNet.nodes[self.id]['t_MI'] >= self.SIRSNet.nodes[self.id]['t_MI_C']:
                     self.SIRSState = NetworkSimulation.CONTAINED         
                     self.SIRSNet.nodes[self.id]['t_C'] = self.sim.now()
                     
        elif self.SIRSState == NetworkSimulation.ACUTE_INFECTED:
            if self.sim.now() - self.SIRSNet.nodes[self.id]['t_AI'] == self.SIRSNet.nodes[self.id]['t_AI_C'] :
                self.SIRSState = NetworkSimulation.CONTAINED         
                self.SIRSNet.nodes[self.id]['t_C'] = self.sim.now()
                

                                                                  

    def mayBecomeRecoveredOrDead(self):
      
        if self.SIRSNet.nodes[self.id]['t_MI'] != 0:
            if self.sim.now() - self.SIRSNet.nodes[self.id]['t_C'] >= self.SIRSNet.nodes[self.id]['t_C_R/MI'] :
                self.SIRSState = NetworkSimulation.RECOVERED_MI
                    
        elif self.SIRSNet.nodes[self.id]['t_AI'] != 0:
            if SIRSAgent.r.random() <= P_AI_R:  
                if self.sim.now() - self.SIRSNet.nodes[self.id]['t_C'] >= self.SIRSNet.nodes[self.id]['t_C_R/AI'] : 
                    self.SIRSState = NetworkSimulation.RECOVERED_AI
            else:
                self.SIRSState = NetworkSimulation.DEAD
            
                    
                
    def mayStayRecoveredOrDead(self):
        if self.SIRSState == NetworkSimulation.RECOVERED_MI:
               self.SIRSState = NetworkSimulation.RECOVERED_MI
        elif self.SIRSState == NetworkSimulation.RECOVERED_AI:
               self.SIRSState = NetworkSimulation.RECOVERED_AI
      
            
        
            
                         
